# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

__metaclass__ = type


class ExtensionModule(BaseJBossModule):
    def __init__(self, context=None):
        super(ExtensionModule, self).__init__(path='/extension', context=context)

    def apply(self, extension=None, **kwargs):
        """
        Apply an extension operation:

        Example:
        extension = [
                {
                    'name': 'org.keycloak.keycloak-adapter-subsystem',
                    'state': 'present',
                    'module': 'org.keycloak.keycloak-adapter-subsystem'
                },
                {
                    'name': 'org.keycloak.keycloak-adapter-subsystem',
                    'state': 'absent'
                }
            ]

        :param extension: a list or single dict of an extension
        :param kwargs: any other args that may be useful
        :return: changed flag and a list of changes that have been applied
        """

        extension = self._format_apply_param(extension)

        changes = []

        for ext in extension:

            state = self._get_param(ext, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('Extension state is not one of [present|absent]')

            name = self._get_param(ext, 'name')

            if state == 'present':
                module = self._get_param(ext, 'module', None)
                changes += self.apply_extension_present(name, module)
            elif state == 'absent':
                changes += self.apply_extension_absent(name)

        return None if len(changes) < 1 else changes

    def apply_extension_absent(self, name):
        try:
            self.cmd('%s=%s:remove()' % (self.path, name))
            return [{'extension': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_extension_present(self, name, module):
        try:
            ext = self.read_resource('%s=%s' % (self.path, name))
            if ext.get('module', None) != module:
                raise ParameterError('WFLYCTL0048: Attribute module is not writable')
            else:
                return []
        except NotFoundError:
            self.cmd('%s=%s:add(module=%s)' % (self.path, name, module))
            return [{'extension': name, 'action': 'add'}]



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

try:
    import simplejson as json
except ImportError:
    import json


class SocketBindingModule(BaseJBossModule):
    BINDING_PARAMS = [
        'port',
        'interface'
    ]

    def __init__(self, context=None):
        super(SocketBindingModule, self).__init__(path='/socket-binding-group=%s/socket-binding=%s',
                                                  context=context)

    def __call__(self, binding_conf):

        binding_conf = self.unescape_keys(binding_conf)

        if 'name' not in binding_conf:
            raise ParameterError('provided socket binding name is null')

        if 'socket-binding-group-name' not in binding_conf:
            raise ParameterError('provided socket binding group name is null')

    def apply(self, socket_binding=None, **kwargs):

        socket_bindings = self._format_apply_param(socket_binding)

        changes = []

        for binding in socket_bindings:

            state = self._get_param(binding, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('socket binding state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_socket_binding_present(binding)
            elif state == 'absent':
                changes += self.apply_socket_binding_absent(binding)

        return None if len(changes) < 1 else changes

    def apply_socket_binding_present(self, binding):
        group_name = self._get_param(binding, 'socket-binding-group-name')
        name = self._get_param(binding, 'name')

        resource_path = self.path % (group_name, name)

        changes = []

        try:
            binding_dmr = self.read_resource_dmr(resource_path, True)
            fc = dict(
                (k, v) for (k, v) in iteritems(binding) if k in self.BINDING_PARAMS)
            a_changes = self._sync_attributes(parent_node=binding_dmr,
                                              parent_path=resource_path,
                                              target_state=fc,
                                              allowable_attributes=self.BINDING_PARAMS)
            if len(a_changes) > 0:
                changes.append({'socket-binding': name, 'socket-binding-group': group_name, 'action': 'update',
                                'changes': a_changes})

        except NotFoundError:
            binding_params = self.convert_to_dmr_params(binding, self.BINDING_PARAMS)

            self.cmd('%s:add(%s)' % (resource_path, binding_params))
            changes.append({'socket-binding': name, 'socket-binding-group': group_name, 'action': 'add',
                            'params': binding_params})

        return changes

    def apply_socket_binding_absent(self, binding):
        group_name = self._get_param(binding, 'socket-binding-group-name')
        name = self._get_param(binding, 'name')
        resource_path = self.path % (group_name, name)
        try:
            self.cmd('%s:remove' % resource_path)
            return [{'socket-binding': name, 'socket-binding-group': group_name, 'action': 'delete'}]
        except NotFoundError:
            return []



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

__metaclass__ = type


class EEModule(BaseJBossModule):
    EE_PARAMS = [
        'service'
    ]

    EE_SERVICE_PARAMS = [
        'datasource'
    ]

    def __init__(self, context=None):
        super(EEModule, self).__init__(path='/subsystem=ee', context=context)

    def apply(self, ee=None, **kwargs):
        """
        Apply an ee subsystem operation:

        Example:

        """
        if not isinstance(ee, dict):
            raise ParameterError('%s provided to %s is not an allowable type' % (ee, self.__class__.__name__))

        changes = []

        for key in ee.keys():
            if key in self.EE_PARAMS and key == 'service':
                services = self._format_apply_param(ee['service'])
                changes += self.apply_ee_service(services)
                # handle service

        return changes if len(changes) > 0 else None

    def apply_ee_service(self, services):

        changes = []
        for service in services:
            name = self._get_param(service, 'name')
            try:
                service_dmr = self.read_resource_dmr('%s/service=%s' % (self.path, name))
                fc = dict(
                    (k, v) for (k, v) in iteritems(service) if k in self.EE_SERVICE_PARAMS)
                a_changes = self._sync_attributes(parent_node=service_dmr,
                                                  parent_path='%s/service=%s' % (self.path, name),
                                                  target_state=fc,
                                                  allowable_attributes=self.EE_SERVICE_PARAMS)
                if len(a_changes) > 0:
                    changes.append({'service': name, 'action': 'update', 'changes': a_changes})

            except NotFoundError as e:
                raise ParameterError('Failed to update EE default bindings: %s' % e.message)

        return changes



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

__metaclass__ = type

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()


class UndertowModule(BaseJBossModule):
    def __init__(self, context=None):
        super(UndertowModule, self).__init__(path='/subsystem=undertow', context=context)

        self.UNDERTOW_PARAMS = {
            "server_name",
            "host_name"
        }

        self.CONFIG_SUBMODULE = {
            'filter_ref': UndertowFilterRefModule(self.context),
            'web_filter': UndertowFilterModule(self.context),
            'http_listener': UndertowHttpListenerModule(self.context),
            'ajp_listener': UndertowAjpListenerModule(self.context)
        }

    def apply(self, undertow=None, **kwargs):

        changes = []

        undertow_config = dict((k, v) for (k, v) in undertow.items() if k not in self.CONFIG_SUBMODULE)

        undertow_config.update(kwargs)

        for key in undertow.keys():
            if key in self.CONFIG_SUBMODULE:
                handler = self.CONFIG_SUBMODULE[key]
                subchanges = handler.apply(undertow[key], **undertow_config)
                if subchanges is not None:
                    changes.append({
                        key: subchanges
                    })
            elif key in self.UNDERTOW_PARAMS:
                pass
            else:
                raise ParameterError('%s cannot handle configuration %s' % (self.__class__.__name__, key))

        return None if len(changes) < 1 else changes


class UndertowFilterRefModule(BaseJBossModule):
    DEFAULT_PRIORITY = 1

    FILTER_REF_PARAMS = {'priority'}

    def __init__(self, context=None):
        super(UndertowFilterRefModule, self).__init__(path='/subsystem=undertow/server=%s/host=%s/filter-ref=%s',
                                                      context=context)

    def apply(self, filter_ref=None, server_name=None, host_name=None, **kwargs):

        filter_refs = self._format_apply_param(filter_ref)

        if server_name is None:
            if 'server_name' in kwargs:
                server_name = kwargs['server_name']
            else:
                raise ParameterError('The undertow filter-ref module requires a server_name argument')

        if host_name is None:
            if 'host_name' in kwargs:
                host_name = kwargs['host_name']
            else:
                raise ParameterError('The undertow filter-ref module requires a host_name argument')

        changes = []

        for fref in filter_refs:

            state = self._get_param(fref, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('filter state is not one of [present|absent]')

            name = self._get_param(fref, 'name')

            if state == 'present':
                changes += self.apply_filter_ref_present(server_name, host_name, name, fref)

            elif state == 'absent':
                changes += self.apply_filter_ref_absent(server_name, host_name, name)

        return None if len(changes) < 1 else changes

    def apply_filter_ref_absent(self, server_name, host_name, name):

        changes = []
        try:
            self.cmd(self.path % (server_name, host_name, name) + ':remove()')
            changes.append({'filter_ref': name, 'action': 'delete'})
        except NotFoundError:
            pass

        return changes

    def apply_filter_ref_present(self, server_name, host_name, name, filter_config):
        changes = []
        filter_ref_path = self.path % (server_name, host_name, name)
        try:
            dmr_filter_ref = self.read_resource_dmr(filter_ref_path)
            # reduce filter config to only allowable values
            fc = dict(
                (key, value) for (key, value) in filter_config.items() if key in self.FILTER_REF_PARAMS)

            a_changes = self._sync_attributes(parent_node=dmr_filter_ref,
                                              parent_path=filter_ref_path,
                                              target_state=fc,
                                              allowable_attributes=self.FILTER_REF_PARAMS,
                                              callback_handler=self._update_filter_ref_attribute)

            if len(a_changes) > 0:
                changes.append({'filter-ref': name, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            # filter ref does not exist and we need to create it
            priority = filter_config.get('priority', None)
            if priority is None or priority == self.DEFAULT_PRIORITY:
                cmd = '/subsystem=undertow/server=%s/host=%s/filter-ref=%s:add' % (server_name, host_name, name)
            else:
                cmd = '/subsystem=undertow/server=%s/host=%s/filter-ref=%s:add(priority=%s)' % (
                    server_name, host_name, name, priority)

            self.cmd(cmd)
            changes.append({'filter-ref': name, 'action': 'add'})

        return changes

    def _update_filter_ref_attribute(self, parent_path=None, name=None, old_value=None, new_value=None):

        change = None

        if name == 'priority' and (old_value is None or old_value == self.DEFAULT_PRIORITY) and (
                        new_value is None or new_value == self.DEFAULT_PRIORITY):
            pass  # nothing to do priority is already set
        elif new_value is None and old_value is not None:
            self.cmd('%s:undefine-attribute(name=%s)' % (parent_path, name))

            change = {
                'attribute': name,
                'action': 'delete',
                'old_value': old_value
            }

        elif new_value is not None and new_value != old_value:
            self.cmd('%s:write-attribute(name=%s, value=%s)' % (parent_path, name, new_value))

            change = {
                'attribute': name,
                'action': 'update',
                'old_value': old_value,
                'new_value': new_value
            }

        return change


class UndertowFilterModule(BaseJBossModule):
    """
    Module that can modify filters.

    web_filter:
        name: filter-name
        type: type of filter e.g. [custom-filter|response-header]
        state: [present|absent]
        prop_x: property by name from each of the filter type definitions

    """

    FILTER_PARAMS = {
        'custom-filter': [
            'module',
            'class-name'
        ],
        'response-header': [
            'header-name',
            'header-value'
        ]
    }

    def __init__(self, context=None):
        super(UndertowFilterModule, self).__init__(path='/subsystem=undertow/configuration=filter',
                                                   context=context)
        self.filter_ref_module = UndertowFilterRefModule(context)

    def apply(self, web_filter=None, **kwargs):

        if web_filter is None:
            raise ParameterError('filter must not be null')

        filters = self._format_apply_param(web_filter)

        changes = []

        for filter_config in filters:

            state = self._get_param(filter_config, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('filter state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_filter_present(filter_config)
                # the filter configuration carries enough information for us to also create the filter-ref element
                ref_change = self.filter_ref_module.apply(filter_config, **kwargs)
                if ref_change is not None:
                    changes += ref_change
            elif state == 'absent':
                # the filter configuration carries enough information for us to also remove the filter-ref element
                ref_change = self.filter_ref_module.apply(filter_config, **kwargs)
                if ref_change is not None:
                    changes += ref_change
                changes += self.apply_filter_absent(filter_config)

        return None if len(changes) < 1 else changes

    def apply_filter_absent(self, filter_config):
        filter_name = self._get_param(filter_config, 'name')
        filter_type = self._get_param(filter_config, 'type')
        changes = []
        # FIXME if 'WFLYCTL0216' not in failure:
        #     raise ProcessingError('failed to delete filter %s: %s' % (name, failure))
        try:
            self.cmd('%s/%s=%s:remove()' % (self.path, filter_type, filter_name))
            changes.append({'filter': filter_name, 'type': filter_type, 'action': 'delete'})
        except NotFoundError:
            pass

        return changes

    def apply_filter_present(self, filter_config):
        filter_name = self._get_param(filter_config, 'name')
        filter_type = self._get_param(filter_config, 'type')
        changes = []
        try:
            dmr_filter = self.read_resource_dmr('%s/%s=%s' % (self.path, filter_type, filter_name))
            # reduce filter config to only allowable values
            fc = dict(
                (key, value) for (key, value) in filter_config.items() if
                key in self.FILTER_PARAMS.get(filter_type, []))
            a_changes = self._sync_attributes(parent_node=dmr_filter,
                                              parent_path='%s/%s=%s' % (self.path, filter_type, filter_name),
                                              target_state=fc,
                                              allowable_attributes=self.FILTER_PARAMS.get(filter_type, []))

            if len(a_changes) > 0:
                changes.append({'filter': filter_name, 'type': filter_type, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            filter_params = self.convert_to_dmr_params(filter_config, self.FILTER_PARAMS.get(filter_type, []))
            self.cmd('%s/%s=%s:add(%s)' % (self.path, filter_type, filter_name, filter_params))
            changes.append({'filter': filter_name, 'type': filter_type, 'action': 'add', 'params': filter_params})

        return changes


class UndertowListenerModule(BaseJBossModule):
    AJP_LISTENER = 'ajp-listener'
    HTTP_LISTENER = 'http-listener'

    def __init__(self, listener_type, listener_params, context=None):
        super(UndertowListenerModule, self).__init__(path='/subsystem=undertow/server=%s/%s=%s', context=context)
        if listener_type not in [UndertowListenerModule.AJP_LISTENER, UndertowListenerModule.HTTP_LISTENER]:
            raise ParameterError("listener type %s is not one of [%s|%s]" % (
                listener_type, UndertowListenerModule.AJP_LISTENER, UndertowListenerModule.HTTP_LISTENER))
        self.listener_type = listener_type
        self.listener_params = listener_params

    def __call__(self, binding_conf):
        binding_conf = self.unescape_keys(binding_conf)
        if 'name' not in binding_conf:
            raise ParameterError('provided %s name is null' % self.listener_type)

    def apply(self, listener_config=None, server_name=None, **kwargs):

        if server_name is None:
            if 'server_name' in kwargs:
                server_name = kwargs['server_name']
            else:
                raise ParameterError('The undertow %s module requires a server_name argument' % self.listener_type)

        listener_configs = self._format_apply_param(listener_config)

        changes = []

        for listener in listener_configs:

            state = self._get_param(listener, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('%s state is not one of [present|absent]' % self.listener_type)

            if state == 'present':
                changes += self.apply_listener_present(server_name, listener)
            elif state == 'absent':
                changes += self.apply_listener_absent(server_name, listener)

        return None if len(changes) < 1 else changes

    def apply_listener_present(self, server_name, listener):

        debug('%s:add(): %r' % (self.__class__.__name__, listener))
        debug('  >> allowed_params: %r' % self.listener_params)

        name = self._get_param(listener, 'name')

        resource_path = self.path % (server_name, self.listener_type, name)

        changes = []

        try:
            listener_dmr = self.read_resource_dmr(resource_path, True)
            fc = dict(
                (k, v) for (k, v) in iteritems(listener) if k in self.listener_params)
            a_changes = self._sync_attributes(parent_node=listener_dmr,
                                              parent_path=resource_path,
                                              target_state=fc,
                                              allowable_attributes=self.listener_params)
            if len(a_changes) > 0:
                changes.append({self.listener_type: name, 'action': 'update',
                                'changes': a_changes})

        except NotFoundError:
            listener_add_params = self.convert_to_dmr_params(listener, self.listener_params)
            self.cmd('%s:add(%s)' % (resource_path, listener_add_params))
            changes.append({self.listener_type: name, 'action': 'add',
                            'params': listener_add_params})

        return changes

    def apply_listener_absent(self, server_name, listener):

        name = self._get_param(listener, 'name')

        resource_path = self.path % (server_name, self.listener_type, name)

        try:
            self.cmd('%s:remove' % resource_path)
            return [{self.listener_type: name, 'action': 'delete'}]
        except NotFoundError:
            return []


class UndertowAjpListenerModule(UndertowListenerModule):
    LISTENER_PARAMS = ["allow-encoded-slash",
                       "allow-equals-in-cookie-value",
                       "always-set-keep-alive",
                       "buffer-pipelined-data",
                       "buffer-pool",
                       "decode-url",
                       "disallowed-methods",
                       "enabled",
                       "max-buffered-request-size",
                       "max-connections",
                       "max-cookies",
                       "max-header-size",
                       "max-headers",
                       "max-parameters",
                       "max-post-size",
                       "no-request-timeout",
                       "read-timeout",
                       "receive-buffer",
                       "record-request-start-time",
                       "redirect-socket",
                       "request-parse-timeout",
                       "resolve-peer-address",
                       "proxy-address-forwarding",
                       "scheme",
                       "secure",
                       "send-buffer",
                       "socket-binding",
                       "tcp-backlog",
                       "tcp-keep-alive",
                       "url-charset",
                       "worker",
                       "write-timeout"]

    def __init__(self, context=None):
        super(UndertowAjpListenerModule, self).__init__(listener_type=UndertowListenerModule.AJP_LISTENER,
                                                        listener_params=UndertowAjpListenerModule.LISTENER_PARAMS,
                                                        context=context)

    def apply(self, ajp_listener=None, **kwargs):
        return super(UndertowAjpListenerModule, self).apply(ajp_listener, **kwargs)


class UndertowHttpListenerModule(UndertowListenerModule):
    LISTENER_PARAMS = ["allow-encoded-slash",
                       "allow-equals-in-cookie-value",
                       "always-set-keep-alive",
                       "buffer-pipelined-data",
                       "buffer-pool",
                       "decode-url",
                       "disallowed-methods",
                       "enabled",
                       "max-buffered-request-size",
                       "max-connections",
                       "max-cookies",
                       "max-header-size",
                       "max-headers",
                       "max-parameters",
                       "max-post-size",
                       "no-request-timeout",
                       "read-timeout",
                       "receive-buffer",
                       "record-request-start-time",
                       "redirect-socket",
                       "request-parse-timeout",
                       "resolve-peer-address",
                       "proxy-address-forwarding",
                       "scheme",
                       "secure",
                       "send-buffer",
                       "socket-binding",
                       "tcp-backlog",
                       "tcp-keep-alive",
                       "url-charset",
                       "worker",
                       "write-timeout"]

    def __init__(self, context=None):
        super(UndertowHttpListenerModule, self).__init__(listener_type=UndertowListenerModule.HTTP_LISTENER,
                                                         listener_params=UndertowAjpListenerModule.LISTENER_PARAMS,
                                                         context=context)

    def apply(self, http_listener=None, **kwargs):
        return super(UndertowHttpListenerModule, self).apply(http_listener, **kwargs)



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()


class DatasourcesModule(BaseJBossModule):
    SUBSYSTEM_PARAMS = [
        'data-source',
        'jdbc-driver'
    ]

    DATASOURCE_PARAMS = [
        'allocation-retry',
        'allocation-retry-wait-millis',
        'allow-multiple-users',
        'background-validation',
        'background-validation-millis',
        'blocking-timeout-wait-millis',
        'capacity-decrementer-class',
        'capacity-decrementer-properties',
        'capacity-incrementer-class',
        'capacity-incrementer-properties',
        'check-valid-connection-sql',
        'connectable',
        'connection-listener-class',
        'connection-listener-property',
        'connection-url',
        'datasource-class',
        'driver-class',
        'driver-name',
        'enabled',
        'enlistment-trace',
        'exception-sorter-class-name',
        'exception-sorter-properties',
        'flush-strategy',
        'idle-timeout-minutes',
        'initial-pool-size',
        'jndi-name',
        'jta',
        'max-pool-size',
        'mcp',
        'min-pool-size',
        'new-connection-sql',
        'password',
        # 'pool-name', FIXME check out why this is not writable
        'pool-fair',
        'pool-prefill',
        'pool-use-strict-min',
        'prepared-statements-cache-size',
        'query-timeout',
        'reauth-plugin-class-name',
        'reauth-plugin-properties',
        'security-domain',
        'set-tx-query-timeout',
        'share-prepared-statements',
        'spy',
        'stale-connection-checker-class-name',
        'stale-connection-checker-properties',
        'statistics-enabled',
        'track-statements',
        'tracking',
        'transaction-isolation',
        'url-delimiter',
        'url-selector-strategy-class-name',
        'use-ccm',
        'use-fast-fail',
        'use-java-context',
        'use-try-lock',
        'user-name',
        'valid-connection-checker-class-name',
        'valid-connection-checker-properties',
        'validate-on-match'
    ]

    JDBC_DRIVER_PARAMS = [
        'deployment-name',
        'driver-class-name',
        'driver-datasource-class-name',
        'driver-major-version',
        'driver-minor-version',
        'driver-module-name',
        'driver-name',
        'driver-xa-datasource-class-name',
        'module-slot',
        'profile',
        'xa-datasource-class'
    ]

    JDBC_DRIVER_NON_UPDATEABLE_PARAMS = [
        'deployment-name',
        'driver-class-name',
        'driver-datasource-class-name',
        'driver-xa-datasource-class-name',
        'driver-major-version',
        'driver-minor-version',
        'driver-module-name',
        'driver-name',
        'module-slot',
        'profile',
        'xa-datasource-class'
    ]

    def __init__(self, context=None):
        super(DatasourcesModule, self).__init__(path='/subsystem=datasources', context=context)

    def apply(self, datasources=None, **kwargs):

        datasources = self.unescape_keys(datasources)

        if not isinstance(datasources, dict):
            raise ParameterError('%s provided to %s is not an allowable type' % (datasources, self.__class__.__name__))

        changes = []

        for key in datasources.keys():

            if key == 'data-source':
                ds = self._format_apply_param(datasources[key])
                changes += self.apply_datasources(ds)
            elif key == 'jdbc-driver':
                drivers = self._format_apply_param(datasources[key])
                changes += self.apply_jdbc_drivers(drivers)
            else:
                debug('%s.apply(): Unrecognised property %s => %r' % (self.__class__.__name__, key, datasources[key]))

        return changes if len(changes) > 0 else None

    def apply_datasources(self, datasources):

        changes = []
        for ds in datasources:

            name = self._get_param(ds, 'name')

            state = self._get_param(ds, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('The datasource state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_datasource_present(name, ds)
            elif state == 'absent':
                changes += self.apply_datasource_absent(name)

        return changes

    def apply_datasource_absent(self, name):

        try:
            self.cmd('%s/data-source=%s:remove()' % (self.path, name))
            return [{'datasource': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_datasource_present(self, name, datasource):

        ds_path = '%s/data-source=%s' % (self.path, name)

        changes = []
        try:
            ds_dmr = self.read_resource_dmr(ds_path)
            # update datasource
            fc = dict(
                (k, v) for (k, v) in iteritems(datasource) if k in self.DATASOURCE_PARAMS)
            a_changes = self._sync_attributes(parent_node=ds_dmr,
                                              parent_path=ds_path,
                                              target_state=fc,
                                              allowable_attributes=self.DATASOURCE_PARAMS)
            if len(a_changes) > 0:
                changes.append({'datasource': name, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            # create the datasource
            ds_params = self.convert_to_dmr_params(datasource, self.DATASOURCE_PARAMS)
            self.cmd('%s/data-source=%s:add(%s)' % (self.path, name, ds_params))
            changes.append({'datasource': name, 'action': 'add', 'params': ds_params})

        return changes

    # JDBC Driver
    def apply_jdbc_drivers(self, jdbc_drivers):

        changes = []
        for jdbc_driver in jdbc_drivers:

            name = self._get_param(jdbc_driver, 'name')

            state = self._get_param(jdbc_driver, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('The jdbc driver state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_jdbc_driver_present(name, jdbc_driver)
            elif state == 'absent':
                changes += self.apply_jdbc_driver_absent(name)

        return changes

    def apply_jdbc_driver_absent(self, name):

        try:
            self.cmd('%s/jdbc-driver=%s:remove()' % (self.path, name))
            return [{'jdbc-driver': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_jdbc_driver_present(self, name, jdbc_driver):

        jdbc_driver_path = '%s/jdbc-driver=%s' % (self.path, name)

        changes = []
        try:
            jdbc_driver_dmr = self.read_resource_dmr(jdbc_driver_path)
            # update jdbc_driver
            for (k, v) in iteritems(jdbc_driver):
                if k in self.JDBC_DRIVER_NON_UPDATEABLE_PARAMS:
                    debug('Warning, parameter %s cannot be updated on jdbc driver and will be ignored' % k)

            fc = dict(
                (k, v) for (k, v) in iteritems(jdbc_driver) if k in self.JDBC_DRIVER_PARAMS and k not in self.JDBC_DRIVER_NON_UPDATEABLE_PARAMS)
            a_changes = self._sync_attributes(parent_node=jdbc_driver_dmr,
                                              parent_path=jdbc_driver_path,
                                              target_state=fc,
                                              allowable_attributes=self.JDBC_DRIVER_PARAMS)
            if len(a_changes) > 0:
                changes.append({'jdbc-driver': name, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            # little hack to maintain sanity, for some reason one has to supply the name of the driver in the list of
            # parameters again, the param itself appears to be thrown away however
            if 'driver-name' not in jdbc_driver:
                jdbc_driver['driver-name'] = name

            jdbc_driver_params = self.convert_to_dmr_params(jdbc_driver, self.JDBC_DRIVER_PARAMS)
            self.cmd('%s/jdbc-driver=%s:add(%s)' % (self.path, name, jdbc_driver_params))
            changes.append({'jdbc-driver': name, 'action': 'add', 'params': jdbc_driver_params})

        return changes



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import time
from functools import wraps
import collections
import os
from abc import ABCMeta, abstractmethod
from synchronize import make_synchronized

from jyboss.exceptions import ContextError, ConnectionError
from jyboss.logging import debug, warn, SyslogOutputStream

try:
    from java.lang import IllegalStateException, IllegalArgumentException, System, ClassLoader, Thread
    from java.io import PrintStream, File
    from java.net import URL, URLClassLoader
    from jarray import array
except ImportError as jpe:
    raise ContextError('Java packages are not available, please run this module with jython.', jpe)

# Python2 & 3 way to get NoneType
NoneType = type(None)

__metaclass__ = type

MODE_EMBEDDED = 'embedded'

MODE_STANDALONE = 'standalone'

streams = collections.namedtuple('streams', ['out', 'err'])


# helper functions
def normalize_dirpath(dirpath):
    return dirpath[:-1] if dirpath[-1] == '/' or dirpath[-1] == '\\' else dirpath


def retry(exceptions=None, tries=None, delay=2, backoff=2):
    if exceptions:
        exceptions = tuple(exceptions)

    def wrapper(fun):

        @wraps(fun)
        def retry_calls(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return fun(*args, **kwargs)
                except exceptions as e:
                    wrapped = ''
                    if len(args) > 0:
                        wrapped = args[0].__class__.__name__
                    if fun is not None and hasattr(fun, 'func_name'):
                        wrapped = ''.join([wrapped, '.', getattr(fun, 'func_name')])
                    warn('%s: %s, Retrying in %d seconds...' % (wrapped, str(e), mdelay))
                    time.sleep(mdelay)
                    mtries -= 1
                    if backoff > 0:
                        mdelay *= backoff
            return fun(*args, **kwargs)

        return retry_calls  # decorator

    return wrapper


class ConfigurationChangeHandler(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def configuration_changed(self, change_event):
        pass


class ConnectionResource(object):
    def __init__(self, mode, context=None):
        if mode is None:
            ContextError('%s cannot be created without a connection mode' % self.__class__.__name__)

        self.mode = mode

        if context is None:
            self.context = JyBossContext.instance()
        else:
            self.context = context

        self.connection = None  # type: Connection

    def __enter__(self):

        if self.mode == MODE_EMBEDDED:
            self.connection = EmbeddedConnection(self.context)
        elif self.mode == MODE_STANDALONE:
            self.connection = ServerConnection(self.context)

        self.connection.connect()

        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        connection = self.connection
        connection.disconnect()
        return False

    def connect(self):
        self.__enter__()

    def disconnect(self):
        self.__exit__(None, None, None)


class Connection(ConfigurationChangeHandler):
    """
    abstract connection to a jboss server
    """
    __metaclass__ = ABCMeta

    def __init__(self, context=None):
        if context is None:
            self.context = JyBossContext.instance()
        else:
            self.context = context
        # need to register to receive change notifications from the context so we can detect if a connection is in progress
        self.context.register_change_handler(self)
        self.context.connection = self
        self.jcli = None

    @abstractmethod
    def _connect(self, cli):
        """Method documentation"""
        return

    @abstractmethod
    def get_mode(self):
        """Method documentation"""
        return

    def connect(self):

        # now we are interested in context change events
        self.context.register_change_handler(self)

        if self.jcli is None:
            jcli = self.context.create_cli()
        else:
            if self.jcli.is_connected():
                raise ContextError('%s.connect: this resource is already connected' % self.__class__.__name__)
            else:
                jcli = self.jcli

        debug("%s.connect: try to connect, current cli: %s" % (self.__class__.__name__, jcli))
        try:
            self._connect(jcli)
            debug("%s.connect: connected to server" % self.__class__.__name__)
        except Exception as e:
            if e.message.startswith('Already connected to server'):
                debug('%s.connect: already connected to server' % self.__class__.__name__)
            else:
                try:  # cleanup and try again
                    jcli.disconnect()
                except:
                    pass
                raise e

        self.jcli = jcli

    def disconnect(self):
        debug("Session.disconnect: try to disconnect %s cli state %s" % (self.get_mode(), self.jcli))
        if self.jcli is not None:
            try:
                self.jcli.disconnect()
            except Exception as e:
                warn('failed to disconnect from %s server: %s' % (self.__class__.__name__, e.message))
                pass
            finally:
                self.jcli = None
                self.context.unregister_change_handler(self)

    def configuration_changed(self, change):
        if 'interactive' in change and self.jcli is not None:
            debug('%s handled: %r' % (self.__class__.__name__, change))
            # TODO self.jcli.silent = change['interactive']['new_value']


class EmbeddedConnection(Connection):
    """
    a cli object that can start an embedded jboss
    """

    def __init__(self, context=None):
        super(EmbeddedConnection, self).__init__(context=context)

    def _connect(self, cli):
        config_file = self.context.config_file if self.context.config_file is not None else 'standalone.xml'

        cli.embedded(self.context.get_jboss_home(), config_file)

    def get_mode(self):
        return 'embedded'


class ServerConnection(Connection):
    """
    a cli object that can interact with jboss server
    """

    def __init__(self, protocol=None, controller_host=None, controller_port=None, admin_username=None,
                 admin_password=None,
                 context=None):
        super(ServerConnection, self).__init__(context=context)
        self._protocol = protocol
        self._controller_host = controller_host
        self._controller_port = controller_port
        self._admin_username = admin_username
        self._admin_password = admin_password

    @retry([ConnectionError], tries=4, backoff=2)
    def _connect(self, cli):
        try:
            cli.connect(protocol=self._protocol,
                        controller_host=self._controller_host,
                        controller_port=self._controller_port,
                        username=self._admin_username,
                        password=self._admin_password)
            debug('%s.connect: connected to server' % self.__class__.__name__)
        except IllegalStateException as e:
            if e.message.startswith('Already connected to server'):
                debug('%s.connect: already connected to server' % self.__class__.__name__)
            else:
                try:  # cleanup and try again
                    cli.disconnect()
                except:
                    pass
                raise ConnectionError(e)

    def get_mode(self):
        return 'standalone'


class ConfigurationChangeObserver(object):
    def __init__(self):
        self.test = None
        self._handlers = []

    def register(self, handler):
        """ Register a handler that will get notified when the configuration in context changes.
        :param handler: the handler to be notified
        :return: True if the handler was appended or False if the handler to be appended was already present
        """
        if not isinstance(handler, ConfigurationChangeHandler):
            raise ContextError('Cannot register handler of type %r' % type(handler))
        elif not any(x == handler for x in self._handlers):
            self._handlers.append(handler)
            return True
        else:
            return False

    def handlers(self):
        return self._handlers

    def unregister(self, handler):
        self._handlers.remove(handler)

    def publish(self, change):
        """ This function will call handler.handle(connection) on every registered handler.
        :param change: the change
        :return: nothing
        """
        debug('%s.publish: %r' % (self.__class__.__name__, change))
        for handler in self._handlers:
            handler.configuration_changed(change)


class JyBossContext(ConfigurationChangeHandler):
    """
    The JyBoss context which manages the bootstrapping of the cli and connection in flight
    """
    _DEFAULT_INSTANCE = None

    _CliType = None

    _EXCLUDE_FROM_OBSERVATION = [
        'change_observer',
        'original_streams',
        'silent_streams',
        '_jboss_home_classpath'
    ]

    def __init__(self, jboss_home=None, config_file=None, interactive=True):
        self.jboss_home = jboss_home
        self.config_file = config_file
        self.original_streams = streams(System.out, System.err)
        self.silent_streams = None
        self.interactive = interactive
        self.connection = None
        # TODO save original streams before nuking them
        self._jboss_home_classpath = None
        # add myself to the list of handlers so context can handle some of the updates to itself
        change_observer = ConfigurationChangeObserver()
        change_observer.register(self)
        self.change_observer = change_observer

    def __setattr__(self, name, value):
        """
        The context item will have a little magic so we can set properties and still be able to notify the change
        handlers to make adjustments should they need to.
        :param name: the name of the context parameter that is changed
        :param value: the value it is changed to
        :return:
        """
        if name not in self._EXCLUDE_FROM_OBSERVATION \
                and name in self.__dict__ \
                and hasattr(self, 'change_observer'):
            old_value = self.__dict__[name]
            if old_value != value:
                self.change_observer.publish({name: {'old_value': old_value, 'new_value': value}})

        super(JyBossContext, self).__setattr__(name, value)

    @staticmethod
    def instance():

        if JyBossContext._DEFAULT_INSTANCE is None:
            JyBossContext._DEFAULT_INSTANCE = JyBossContext()

        return JyBossContext._DEFAULT_INSTANCE

    def create_cli(self):
        # make sure we only ever load this once
        if JyBossContext._CliType is None:
            JyBossContext._CliType = self._load_cli()

        return JyBossContext._CliType()

    @make_synchronized
    def _set_interactive(self, interactive):
        # deactivate all handlers
        if not interactive:
            debug("disable default JVM output streams")
            if self.silent_streams is None:
                self.silent_streams = streams(SyslogOutputStream(), SyslogOutputStream(SyslogOutputStream.ERR))
            System.out.flush()
            System.err.flush()
            System.setOut(PrintStream(self.silent_streams.out, True))
            System.setErr(PrintStream(self.silent_streams.err, True))
        else:
            debug("enable default JVM output streams")
            System.out.flush()
            System.err.flush()
            System.setOut(self.original_streams.out)
            System.setErr(self.original_streams.err)

        # FIXME boadcast configuration change to connection managers

        if self.connection is not None and self.connection.jcli is not None:
            self.connection.jcli.set_silent(not interactive)

    def get_jboss_home(self):

        jboss_home = self.jboss_home

        # first check if the jboss home was passed in as java property
        if jboss_home is None:
            jboss_home = System.getProperty('jboss.home.dir')

        # if still nothing, try to check the os environment JBOSS_HOME
        if jboss_home is None:
            jboss_home = os.environ.get("JBOSS_HOME")

        if jboss_home is None:
            raise ContextError("jboss_home must be provided to the module context")

        # preserver for later
        self.jboss_home = jboss_home

        return jboss_home

    def is_connected(self):
        """ Check if a context is connected
        :return: True if a connection is available, False otherwise
        """
        return self.connection is not None and self.connection.jcli is not None and self.connection.jcli.is_connected()

    def disconnect(self):
        if self.is_connected():
            self.connection.disconnect()
        else:
            raise ContextError("no session in progress")

    def _load_cli(self):
        try:
            # @formatter:off
            # noinspection PyUnresolvedReferences
            from org.jboss.logmanager import LogManager as JBossLogManager
            from org.jboss.as.cli import CommandContextFactory
            from org.jboss.as.cli import CliInitializationException
            from org.jboss.as.cli import CommandLineException
            # @formatter:on
        except ImportError:
            self._configure_classpath()
            try:
                # @formatter:off
                # noinspection PyUnresolvedReferences
                from org.jboss.logmanager import LogManager as JBossLogManager
                from org.jboss.as.cli import CommandContextFactory
                from org.jboss.as.cli import CliInitializationException
                from org.jboss.as.cli import CommandLineException
                # @formatter:on
            except ImportError:
                raise ContextError(
                    "jboss cli libraries are not on the classpath, either start jython with jboss-cli-client.jar on the classpath, set JBOSS_HOME environment variable or create a context with a specific jboss_home path")

        # if the log manager is JUL we need to hack it as it has been initialised prior we got a change to do so,
        # embedded mode will fail if this is not setup properly
        try:
            from java.util.logging import LogManager as JulLogManager
            from java.lang.reflect import Modifier
        except ImportError as jpe:
            raise ContextError('Java packages are not available, please run this module with jython.', jpe)

        logManager = JulLogManager.getLogManager()
        if type(logManager) is JulLogManager:
            # need to hack the JUL LogManager
            field = logManager.__class__.getDeclaredField("manager")
            field.setAccessible(True)
            modifiersField = field.__class__.getDeclaredField("modifiers")
            modifiersField.setAccessible(True)
            modifiersField.setInt(field, field.getModifiers() & ~Modifier.FINAL)
            field.set(None, JBossLogManager())

        # now it's safe to load the cli
        from .cli import Cli as _Cli
        return _Cli

    def _configure_classpath(self):
        """
        jboss CLI does something dodgy and can't just append the cli.jar to the system path
        sys.path.append(jboss_home + "/bin/client/jboss-cli-client.jar")
        instead we are going to add a URL classloader into the loader hierarchy of
        the current thread context
        """
        jboss_home_str = self.get_jboss_home()

        jboss_home = File(normalize_dirpath(jboss_home_str))
        if not jboss_home.isDirectory():
            raise ContextError("jboss_home %s is not a directory or does not exist" % jboss_home_str)

        jars = array([], URL)
        jars.append(File(jboss_home, "/bin/client/jboss-cli-client.jar").toURL())
        jars.append(File(jboss_home, "/jboss-modules.jar").toURL())
        current_thread_classloader = Thread.currentThread().getContextClassLoader()
        updated_classloader = URLClassLoader(jars, current_thread_classloader)
        Thread.currentThread().setContextClassLoader(updated_classloader)

        self._jboss_home_classpath = jboss_home_str

    def register_change_handler(self, handler):
        self.change_observer.register(handler)

    def unregister_change_handler(self, handler):
        self.change_observer.unregister(handler)

    def configuration_changed(self, change_event):
        if 'interactive' in change_event and 'new_value' in change_event['interactive']:
            self._set_interactive(bool(change_event['interactive']['new_value']))



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from synchronize import make_synchronized
from jyboss.exceptions import ContextError

try:
    from java.io import OutputStream
except ImportError as jpe:
    raise ContextError('Java packages are not available, please run this module with jython.', jpe)

try:
    from org.graylog2.syslog4j import SyslogConstants, Syslog

    HAS_SYSLOG = True
    syslog = Syslog.getInstance("unix_syslog")
    config = syslog.getConfig()
    config.setFacility(SyslogConstants.FACILITY_USER)
    config.setThrowExceptionOnWrite(False)
    config.setIdent(__name__.split(".", 1)[0])
    config.setIncludeIdentInMessageModifier(True)
    # localName ??
except ImportError:
    HAS_SYSLOG = False

__metaclass__ = type


def _log(level, msg):
    if HAS_SYSLOG:
        syslog.log(level, msg)


def emergency(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_EMERGENCY, msg)


def alert(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_ALERT, msg)


def critical(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_CRITICAL, msg)


def error(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_ERROR, msg)


def warn(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_WARN, msg)


def notice(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_NOTICE, msg)


def info(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_INFO, msg)


def debug(msg):
    if HAS_SYSLOG:
        _log(SyslogConstants.LEVEL_DEBUG, msg)


class SyslogOutputStream(OutputStream):
    OUT = "out"
    ERR = "err"

    def __init__(self, type=OUT):
        self.type = type
        self.lines = []

    def write(self, b):
        raise NotImplementedError('SyslogOutputStream.write(byte) is not implemented')

    # have to synchronise this otherwise access to the threads writing
    # to the capture array gets clobbered
    @make_synchronized
    def write(self, b, off, length):
        if off == 0 and length == 1:
            pass
        else:
            if b[length - 1] == 10:
                length -= 1
            line = "".join(map(chr, b[off:length]))
            # also log to syslog as info or errors
            if HAS_SYSLOG:
                if self.type == SyslogOutputStream.OUT:
                    info(line)
                elif self.type == SyslogOutputStream.ERR:
                    error(line)
            self.lines.append(line)



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.context import JyBossContext, MODE_EMBEDDED, MODE_STANDALONE
from jyboss.context import ConnectionResource as _ConnectionResource
from jyboss.command import ls, cmd, cd, batch

#
# from jyboss.command import *

__metaclass__ = type

# make default context available to module user
jyboss = JyBossContext.instance()
standalone = _ConnectionResource(MODE_STANDALONE, jyboss)
embedded = _ConnectionResource(MODE_EMBEDDED, jyboss)
disconnect = jyboss.disconnect



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import sys

__metaclass__ = type


class ConnectionError(Exception):
    """ A connection problem was encountered.
    """

    def __init__(self, cause):
        self.traceback = sys.exc_info()
        self.cause = cause
        super(ConnectionError, self).__init__(cause.message)


class ContextError(Exception):
    """ A problem with a context command is encountered.
    """

    def __init__(self, message, cause=None):
        self.traceback = sys.exc_info()
        self.cause = cause
        super(ContextError, self).__init__(message)


class OperationError(Exception):
    """ A problem with a cli command is encountered.
    """
    pass


class NotFoundError(Exception):
    """ A problem with a cli command is encountered.
    """
    pass


class DuplicateResourceError(Exception):
    """ The resource to be created already exists.
    """
    pass


class ParameterError(Exception):
    """
    error when a wrong parameter was supplied to the module
    """
    pass


class ProcessingError(Exception):
    """
    error processing a module request
    """
    pass


class CommandError(Exception):
    def __init__(self, message, cause=None):
        self.traceback = sys.exc_info()
        self.cause = cause
        super(CommandError, self).__init__(message)



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import NotFoundError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()


class DeploymentModule(BaseJBossModule):
    DEPLOYMENT_PARAMS = [
        'name',
        'enabled'
    ]

    def __init__(self, context=None):
        super(DeploymentModule, self).__init__(path='/deployment=%s', context=context)

    def apply(self, deployment=None, **kwargs):

        deployments = self._format_apply_param(deployment)

        changes = []

        for d in deployments:
            name = self._get_param(d, 'name')
            enabled = self._get_param(d, 'enabled')
            action = 'deploy' if enabled else 'undeploy'
            debug(self.__class__.__name__ + '.apply: ' + (self.path % name) + ':' + action)
            try:
                cur_depl = self.read_resource(self.path % name)
                if 'enabled' in cur_depl:
                    if cur_depl['enabled'] != enabled:
                        self.cmd((self.path % name) + ':' + action)
                        changes.append({'deployment': name, 'action': action})

            except NotFoundError:
                pass

        return changes if len(changes) > 0 else None



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import inspect
import re
import collections
from abc import ABCMeta, abstractmethod
from copy import deepcopy

from jyboss.exceptions import *
from jyboss.logging import debug
from jyboss.context import ConfigurationChangeHandler, JyBossContext

try:
    # noinspection PyUnresolvedReferences
    from java.lang import IllegalArgumentException
    # noinspection PyUnresolvedReferences
    from java.util import NoSuchElementException
except ImportError as jpe:
    raise ContextError('Java packages are not available, please run this module with jython.', jpe)

try:
    import simplejson as json
except ImportError:
    import json

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

# python 2 and 3 way of none type
NoneType = type(None)

__metaclass__ = type


class UndefinedType(object):
    pass


undefined = type(UndefinedType())

_expression_matcher = re.compile('^[\'\"]?(?:expression\s*)?[\'\"]?(.*\$\{.*\}[^\'\"]*)[\'\"]*$')
_not_found_matcher = re.compile('WFLYCTL0030|WFLYCTL0216')


def unescape_keys(d):
    """
    Recursively processes all dictionary keys and replaces '_' with '-' and '#' with '.' . This function is mainly used
    to convert from YAML to jboss format.

    :param d {object} - any type of item to unescape

    :return {object} - the unescaped object with keys all fixed up
    """
    if isinstance(d, dict):
        new = {}
        for k, v in iteritems(d):
            new[k.replace('_', '-')] = unescape_keys(v)
        return new
    elif isinstance(d, list):
        new = []
        for v in d:
            new.append(unescape_keys(v))
        return new
    else:
        return d


def escape_keys(d):
    """
    Recursively processes all dictionary keys and replaces '-' with '_' and '.' with '#' . This function is mainly used
    to convert from jboss to YAML format.

    :param d {object} - any type of item to escape to be valid in YAML

    :return {object} - the escaped object with keys valid for YAML output
    """
    if isinstance(d, dict):
        new = {}
        for k, v in iteritems(d):
            if k == 'EXPRESSION_VALUE':  # collapse
                new = v
            else:
                new[k.replace('-', '_')] = escape_keys(v)
        return new
    elif isinstance(d, list):
        new = []
        for v in d:
            new.append(escape_keys(v))
        return new
    else:
        return d


def expression_deserializer(obj):
    """
    Collapses json nodes that are dicts and have a key name of EXPRESSION_VALUE. This call can be used as an object
    handler on the json.dumps method to collapse any DMR specific expression objects.

    :param obj {object} - the dmr serialised object value to check
    :return {object} - the object or if expression value, the value of the object
    """
    if isinstance(obj, dict) and 'EXPRESSION_VALUE' in obj:
        return obj['EXPRESSION_VALUE']
    else:
        return obj


def converts_to_dmr(obj):
    """
    Converts a python dict or list to a jboss dmr string. This function can be used to generate complex jboss cli
    attribute arguments. See SecurityModule for example use.

    :param obj {dict|list} - The dict or list object to turn into a dmr string
    :return {string} - a DMR string that can be used in the cli commands
    """
    return None if obj is None else json.dumps(obj, separators=(',', '=>'))


def convert_to_dmr_params(args, allowable_attributes=None):
    """
    Converts a dictionary of parameters into a "string" list that can be used in a add method call on the jboss cli).

    :param args {dict} - the argument key/value pair that needs to be turned into a param list
    :param allowable_attributes {list[string]} - a list of attribute names that will be included in the args string
    :return: {string} - a string list of args formatted for the cli
    """
    result = ', '.join(
        ['%s=%s' % (k, convert_type(v)) for k, v in iteritems(args) if
         allowable_attributes is not None and k in allowable_attributes])
    return result


def convert_type(obj):
    """
    Convert an object into a JSON string representation.

    :param obj {object} - the object to convert
    :return {string} - the object as a JSON string
    """
    if isinstance(obj, basestring):
        m = _expression_matcher.match(obj)
        if m:
            return "\"%s\"" % m.group(1)
        else:
            return "\"%s\"" % obj
    if isinstance(obj, dict):
        return '[' + ', '.join(['%s=%s' % (k, convert_type(v)) for (k, v) in iteritems(obj)]) + ']'
    else:
        return json.dumps(obj)


def stringify_object(obj):
    """
    Stringify a python dict so we can better compare it to a JSON serialised DMR object.

    :param obj {object} - the object to stringify
    :return {object} - a stringyfied object of any sort
    """
    if isinstance(obj, dict):
        return dict((k, stringify_object(v)) for k, v in iteritems(obj))
    if isinstance(obj, list):
        return [stringify_object(v) for v in obj]
    if isinstance(obj, bool):
        return unicode(obj).lower()
    elif isinstance(obj, (int, long, basestring)):
        return unicode(obj)
    else:
        raise ParameterError('Cannot convert object value %s of type %s to unicode' % (obj, type(obj)))


def clean_python_value(obj, target_type_hint=None):
    if obj is None:
        return None
    elif isinstance(obj, (int, long, bool)):
        return obj if target_type_hint is None or target_type_hint == NoneType \
            else target_type_hint(obj)
    elif isinstance(obj, basestring):
        try:
            match = _expression_matcher.match(obj).group(1)  # why was expression parser using .search() ?
        except AttributeError:  # no expression wrapping to clean
            match = obj
        return match if target_type_hint is None or target_type_hint == NoneType \
            else target_type_hint(match)
    elif isinstance(obj, list):
        return list(clean_python_value(item) for item in obj)
    elif isinstance(obj, dict):
        return dict((k, clean_python_value(v)) for k, v in iteritems(obj))
    else:
        raise IllegalArgumentException('Value cannot be pre-processed for synching: %r' % obj)


def convert_dmr_to_python(n):
    # TODO barf if not a DMR node
    if n is None \
            or not n.isDefined() \
            or n.type == n.type.UNDEFINED:
        return None
    elif n.type == n.type.INT:
        return n.asInt()
    elif n.type == n.type.LONG:
        return n.asLong()
    elif n.type == n.type.STRING:
        return n.asString()
    elif n.type == n.type.BOOLEAN:
        return n.asBoolean()
    elif n.type == n.type.EXPRESSION:
        exp = n.asExpression()
        return exp.getExpressionString()
    elif n.type == n.type.LIST:
        v_list = []
        for _, v_list_node in enumerate(n.asList()):
            i_v = convert_dmr_to_python(v_list_node)
            v_list.append(i_v)
        return v_list
    elif n.type == n.type.OBJECT:
        props = n.asPropertyList()
        obj = {}
        for prop in props:
            obj[prop.name] = convert_dmr_to_python(prop.value)
        return obj
    else:
        raise IllegalArgumentException('Node type cannot be converted to a python value: %r' % n)


class CommandHandler(ConfigurationChangeHandler):
    def __init__(self, context=None):
        super(CommandHandler, self).__init__()
        if context is None:
            self.context = JyBossContext.instance()
        else:
            self.context = context
        self.escape_keys = escape_keys
        self.unescape_keys = unescape_keys
        self.converts_to_dmr = converts_to_dmr
        self.convert_to_dmr_params = convert_to_dmr_params

    def configuration_changed(self, change):
        debug('%s.handle: cli is %r' % (self.__class__.__name__, change))
        # TODO anything the command handler needs to do on change?

    def _cli(self):
        if not self.context.is_connected():
            raise ContextError('%s: no session in progress, please connect()' % self.__class__.__name__)
        else:
            return self.context.connection.jcli

    def start(self):
        """
        Start a batch command session
        """
        self._cli().batch_start()

    def reset(self):
        """
        Reset batch command session
        """
        self._cli().batch_reset()

    def cmd(self, cmd, silent=False):
        debug('%s.cmd(): %s' % (self.__class__.__name__, cmd))
        result = self._cli().cmd(cmd)
        if result.isSuccess():
            return self._return_success(result, silent=silent)
        else:
            errm = self._extract_errm(result)
            if errm is None:
                raise OperationError('Unknown error occurred executing: %s' % cmd)
            elif _not_found_matcher.match(errm) is not None:
                raise NotFoundError(errm)
            else:
                raise OperationError(errm)

    def cmd_dmr(self, cmd):
        result = self._cli().cmd('%s' % cmd)
        if result.isSuccess():
            r = result.getResponse()
            if r.has('result'):
                return r.get('result')
            else:
                return r
        else:
            errm = self._extract_errm(result)
            if errm is None:
                raise OperationError('Unknown error occurred executing: %s' % cmd)
            elif _not_found_matcher.match(errm) is not None:
                raise NotFoundError(errm)
            else:
                raise OperationError(errm)

    def dmr_to_python(self, node=None):
        if node is None:
            return None
        else:
            node_str = node.toJSONString(True)
            return json.loads(node_str, object_hook=expression_deserializer)

    def _as_value_pair(self, node):
        pass

    def _return_success(self, result, transform_cb=None, silent=False):
        """
        Return an executed response to the caller.

        :param result: the jboss result to transform
        :param transform_cb: a callback method that can transform the result prior to being returned
        :param silent: if the response
        :return: the transformed response
        """
        node = result.getResponse()
        response = self.dmr_to_python(node=node)

        if transform_cb is not None:
            response = transform_cb(response)

        if not self.context.interactive or silent:
            if response is None:
                return 'ok'
            elif 'result' in response:
                return response['result']
            elif 'response' in response:
                return response['response']
            else:
                return response
        else:
            if response is None:
                print('ok')
            elif isinstance(response, dict) and 'result' in response:
                print(json.dumps({'response': response['result']}, indent=4))
            elif isinstance(response, dict) or isinstance(response, list):
                print(json.dumps(response, indent=4))
            else:
                # TODO may want to cater for other types that arrive here ?
                print(repr(response))

    @staticmethod
    def _extract_errm(result, encoding='utf-8'):
        """

        :param result: the JBoss DMR result node
        :return: the error message string
        """
        if result is not None:
            nv = result.getResponse().get('failure-description')
        else:
            nv = None

        nv = nv.asString()
        return None if nv is None else nv.encode(encoding)

    def cd(self, path='.', silent=False):
        try:
            result = self._cli().cmd('cd %s' % path)
            if result.isSuccess():
                return self._return_success(result, silent=silent)
            else:
                raise OperationError(self._extract_errm(result))
        except IllegalArgumentException as e:
            raise OperationError(e.getMessage())

    def ls(self, path=None, silent=False):
        result = self._cli().cmd('ls' if path is None else 'ls %s' % path)
        if result.isSuccess():
            return self._return_success(result, _ls_response_magic, silent=silent)
        else:
            errm = self._extract_errm(result)
            if errm.find('WFLYCTL0062') != -1 and _not_found_matcher.match(errm) is not None:
                # TODO snip errm?
                raise NotFoundError(errm)
            else:
                raise OperationError(errm)

    def run(self, silent=False):
        """
        Run the batch command
        :param silent: if True disable output to stdout
        """
        result = self._cli().cmd('run-batch')
        if result.isSuccess():
            return self._return_success(result, silent=silent)
        else:
            errm = self._extract_errm(result)
            if errm.find('WFLYCTL0062') != -1 and _not_found_matcher.match(errm) is not None:
                raise NotFoundError(errm)
            elif errm.find('WFLYCTL0062') != -1 and errm.find('WFLYCTL0212') != -1:
                raise DuplicateResourceError(errm)
            else:
                raise OperationError(errm)

    def add_cmd(self, batch_cmd):
        """
        Add command to batch command list for execution
        :param batch_cmd: {str|list(str)} - the command to add to the list of batch commands
        """
        if not isinstance(batch_cmd, list):
            batch_cmd = [batch_cmd]

        for cmd in batch_cmd:
            self._cli().batch_add_cmd(cmd)

    def is_active(self):
        return self._cli().batch_is_active()


class ChangeObservable(object):
    """
    Observable class that will be used to process the command line and delegate processing to the actual handler
    """

    def __init__(self):
        self._observers = {}

    def process_instructions(self, instructions):
        actions = []

        local_instructions = deepcopy(instructions)

        for k in instructions.keys():
            instruction = instructions[k]
            if k == 'actions':
                if instruction is None:
                    pass
                elif isinstance(instruction, list):
                    for action in instruction:
                        for ka, va in action.items():
                            actions.append((ka, va))
                else:
                    raise ParameterError('actions must be a list of module instructions')

                # need to remove the key as part of elimination
                del local_instructions[k]

            elif k in self._observers:
                actions.append((k, instruction))
                # need to remove the key as part of elimination
                del local_instructions[k]

        return self._execute_action(local_instructions, actions)

    def _execute_action(self, configuration, actions):
        result = dict(changed=False)

        for key, action in actions:
            debug('jyboss ChangeObservable.process(%s)' % key)
            local_instruction = deepcopy(configuration)
            local_instruction[key] = action
            # notify all observers that can handle this command instruction
            for observer in self._observers.setdefault(key, []):
                changes = observer.apply(**local_instruction)
                if changes is not None:
                    result['changed'] = True
                    # I want result['action'] = key and changes appended to result['changes']
                    result.setdefault('changes', [])
                    result['changes'].append({
                        'module': key,
                        'executor': observer.__class__.__name__,
                        'changes': changes
                    })
                    # FIXME facter needs to return mapped key and no change result[key_replacement_name] = changes

        return result

    def register(self, observer):

        if observer is None:
            raise ParameterError('observer cannot be null')

        elif not hasattr(observer, 'apply'):
            raise NotImplementedError(
                '%s does not have an apply method' % observer.__class__.__name__)

        spec = inspect.getargspec(observer.apply)
        # by definition the first argument in the apply function is the configuration name that can be handled
        if len(spec.args) > 1 and spec.args[0] == 'self':
            key = spec.args[1]
            self._observers.setdefault(key, []).append(observer)
        else:
            raise ParameterError('%s.apply is not implemented correctly' % observer.__class__.__name__)


class ReloadCommandHandler(CommandHandler):
    def __init__(self, context=None):
        super(ReloadCommandHandler, self).__init__(context=context)

    def apply(self, reload=False, **kwargs):
        debug('%s:apply() %r' % (self.__class__.__name__, reload))
        if bool(reload):
            self.cmd('/:reload()')
            return ['reloaded']
        else:
            return None


class CliCmdHandler(CommandHandler):
    """
    A basic handler that will simply execute the cli string given and return. IF an error occurs it will fail
    throw an error and execution halts.

    Example:
        cmd: ''
    """

    def __init__(self, context=None):
        super(CliCmdHandler, self).__init__(context=context)

    def apply(self, cmd=None, **kwargs):
        debug('%s:apply() %r' % (self.__class__.__name__, reload))
        if cmd is not None:
            self.cmd(cmd)

        return None


# TODO review this class
class AttributeUpdateHandler(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def update(self, **kwargs):
        pass


class BaseJBossModule(CommandHandler):
    STATE_ABSENT = 'absent'
    STATE_PRESENT = 'present'

    def __init__(self, path, context=None):
        super(BaseJBossModule, self).__init__(context=context)
        self.path = path
        self.compareLists = lambda x, y: collections.Counter(x) == collections.Counter(y)

    def update_list(self, parent_path=None, name=None, old_value=None, new_value=None, **kwargs):
        change = None
        # check if both lists have the same content (ignoring order or elements)
        if not self.compareLists(old_value, new_value):
            # not the same, clear list and add all values of the new list
            self.cmd('%s:list-clear(name=%s)' % (parent_path, name))
            for i, v in enumerate(new_value):
                self.cmd('%s:list-add(name=%s,index=%s,value="%s")' % (parent_path, name, i, v))
            change = {
                'attribute': name,
                'action': 'update',
                'old_value': old_value,
                'new_value': new_value
            }

        return change

    def update_object(self, parent_path=None, name=None, old_value=None, new_value=None, **kwargs):
        change = None

        if new_value is None and old_value is not None:
            self.cmd('%s:undefine-attribute(name=%s)' % (parent_path, name))

            change = {
                'attribute': name,
                'action': 'delete',
                'old_value': old_value
            }
        elif new_value is not None and old_value is None:
            self.cmd('%s:write-attribute(name=%s, value=%s)' % (parent_path, name, convert_type(new_value)))

            change = {
                'attribute': name,
                'action': 'add',
                'old_value': old_value,
                'new_value': new_value
            }
        else:
            new_value = stringify_object(new_value)
            if new_value != old_value:
                self.cmd('%s:write-attribute(name=%s, value=%s)' % (parent_path, name, convert_type(new_value)))

                change = {
                    'attribute': name,
                    'action': 'update',
                    'old_value': old_value,
                    'new_value': new_value
                }
                # we can't compare objects here since we don't know that their types would be
                # best we can do is turn all fields into string and compare strings

        return change

    def update_attribute(self, parent_path=None, name=None, old_value=None, new_value=None, **kwargs):

        change = None

        if new_value is None and old_value is not None:
            self.cmd('%s:undefine-attribute(name=%s)' % (parent_path, name))

            change = {
                'attribute': name,
                'action': 'delete',
                'old_value': old_value
            }
        elif new_value is not None and new_value != old_value:
            self.cmd('%s:write-attribute(name=%s, value=%s)' % (parent_path, name, convert_type(new_value)))

            change = {
                'attribute': name,
                'action': 'add' if new_value is not None and old_value is None else 'update',
                'old_value': old_value,
                'new_value': new_value
            }

        return change

    def _sync_attributes(self, parent_node=None, parent_path=None, allowable_attributes=None, target_state=None,
                         callback_handler=None, callback_args=None):
        """
        Synchronise the attributes of a configuration object with allowable list of args

        :param parent_node: {ModelNode} - the parent dmr node containing the attributes to sync
        :param parent_path: {string} - the path of the node that needs its attributes synced
        :param allowable_attributes: {list(string)} - a list of attributes that can be updated
        :param target_state: {dict} - the requested target state to sync the parent to
        :param callback_handler: {function} - callback handler that can process the attribute updates
        :param callback_args: {dict} - any other arguments that need to be passed to the callback handler
        :return: changed and changes as list
        """
        if parent_node is None:
            raise ParameterError('A parent node must be provided.')

        # noinspection PyUnresolvedReferences
        from org.jboss.dmr import ModelNode

        if not isinstance(parent_node, ModelNode):
            raise ParameterError('A parent node is not a jboss dmr ModelNode.')

        if parent_path is None:
            raise ParameterError('A parent node path must be provided')

        if target_state is None:
            raise ParameterError('A target state must be provided for the node')

        if callback_handler is None:
            # bind the default callback handler
            callback_handler = self.update_attribute
        elif not inspect.ismethod(callback_handler):  # validate the passed in callback handler
            raise ParameterError('Provided callback_handler is not a function')

        if callback_args is None:
            callback_args = {}

        # add allowable attributes to the callback args if provided else they will be None and ignored by the callback
        callback_args['parent_path'] = parent_path

        changes = []

        for k_t, v_t in target_state.items():

            if allowable_attributes is not None and k_t not in allowable_attributes:
                raise NotImplementedError(
                    'Setting attribute %s is not supported by this module. Node path is %s' % (k_t, parent_path))

            if parent_node.has(k_t):
                attr = parent_node.require(k_t)
            else:
                raise ParameterError('%s.sync_attr: synchronizing attribute %s is not supported on jboss node %s' % (
                    self.__class__.__name__, k_t, parent_path))

            # convert the attribute value to a comparable python value
            try:
                v_a = convert_dmr_to_python(attr)
            except IllegalArgumentException:
                raise ParameterError('%s.sync_attr: synchronizing attribute %s of type %s is not supported' % (
                    self.__class__.__name__, k_t, attr.type))

            # also need to ensure the v_t is escaped/cleaned so we can compare
            v_t = clean_python_value(v_t, type(v_a))

            debug('%s.sync_attr: param %s of type %s will be processed old[%r] new[%r]' % (
                self.__class__.__name__, k_t, attr.type, v_a, v_t))

            callback_args['name'] = k_t
            callback_args['old_value'] = v_a
            callback_args['new_value'] = v_t

            # if either value is of type list, we need to engage a different callback handler as syncing list is
            # not as simple as updating single value attributes, if both are undefined we don't really care and
            # if their types differ, the caster would have already thrown a tantrum
            if isinstance(v_a, list) or isinstance(v_t, list):
                callback_handler = self.update_list

            if isinstance(v_a, dict) or isinstance(v_t, dict):
                callback_handler = self.update_object

            change = callback_handler(**callback_args)
            if change is not None:
                changes.append(change)

        return changes

    def _get_param(self, obj, name, default=undefined):
        """
        Extracts a parameter from the provided configuration object.

        :param obj {dict} - the object to check
        :param name {string} - the name of the param to get
        :param default {any} - will be returned if the object is None or the name is not in the obj
        :return {any} - whatever this param is set to
        """
        if obj is None:
            if default == undefined:
                raise ParameterError('%s: configuration is null' % self.__class__.__name__)
            else:
                return default
        elif name not in obj:
            if default == undefined:
                raise ParameterError('%s: no % s was provided' % (self.__class__.__name__, name))
            else:
                return default
        else:
            return obj[name]

    def read_attribute_dmr(self, resource_path, attribute_name):
        """
        Read a resource and return it as a dmr node

        :param resource_path {string} - the absolute or relative path to the attribute parent from which to read
        :param attribute_name {string} - the attribute name to read
        :return {ModelNode} - a dmr result node
        """
        cmd = '%s:read-attribute(name=%s)' % (resource_path, attribute_name)
        return self.cmd_dmr(cmd)

    def read_resource_dmr(self, resource_path, recursive=False):
        """
        Read a resource and return it as a dmr node

        :param resource_path: {string} - the absolute or relative path to the resource to read
        :param recursive: {bool} - show the resource content recursive
        :return: a dmr node
        """
        cmd = '%s:read-resource(recursive=%s)' % (resource_path, str(recursive).lower())
        return self.cmd_dmr(cmd)

    def read_resource(self, resource_path, recursive=False):
        """
        Read a resource and return it in python type format

        :param resource_path: {string} - the absolute or relative path to the resource to read
        :param recursive: {bool} - show the resource content recursive
        :return: a dmr node
        """
        node = self.read_resource_dmr(resource_path, recursive)
        return self.dmr_to_python(node=node)

    def _format_apply_param(self, arg):
        """
        Apply parameters can be either dicts of list of dicts, this method will simply turn it into a list.
        :param arg: the apply param to format
        :return: a list of apply parameters
        """
        if arg is None:
            result = None
        elif isinstance(arg, dict):
            result = [arg]
        elif isinstance(arg, list):
            result = arg
        else:
            raise ParameterError('%s provided to %s is not an allowable type' % (type(arg), self.__class__.__name__))

        return self.unescape_keys(result)

    @abstractmethod
    def apply(self, **kwargs):
        """
        Method to call with the module configuration to apply the module specific actions.
         :param kwargs {dict} the full configuration set, each module is responsible for picking out
         the bits that it needs
         :return {bool, list} returns a change flag and a list of changes that have been applied
        """
        pass


def _ls_response_magic(response):
    if response is None:
        return None
    elif not isinstance(response, dict):
        return response
    else:
        result = response.get('result', None)
        # if there are steps its attribute and children, else its just the result
        if result is None:
            return None
        elif isinstance(result, list):
            return result
        elif isinstance(result, dict):
            children_step = result.get('step-1')
            attr_step = result.get('step-2')
            nr = dict()
            if children_step is not None:
                nr['children'] = children_step.get('result', None)
            if attr_step is not None:
                nr['attributes'] = attr_step.get('result', None)
            return nr
        else:
            return dict(response=repr(result))
